@@ -15,7 +15,10 @@ from typing import Optional, Callable, Dict, List, Set, Any, Union, MutableMappi
 import functools
 import inspect
 import datetime as dt
-import re
+try:
+    import re2 as re  # type: ignore
+except ModuleNotFoundError:
+    import re  # type: ignore
 import json
 from dataclasses import dataclass
 from enum import Enum
@@ -24,13 +27,11 @@ import flask
 import threading  # for Local and Lock
 
 # for local use & forwarding
-from flask import Response, request
-
-# just for forwarding
 # NOTE the only missing should be "Flask"
-from flask import session, jsonify, Blueprint, make_response, abort, \
-    redirect, url_for, after_this_request, send_file, send_from_directory, \
-    escape, Markup, render_template, current_app, g
+from flask import \
+    Response, Request, request, session, jsonify, Blueprint, make_response, \
+    abort, redirect, url_for, after_this_request, send_file, current_app, g, \
+    send_from_directory, escape, Markup, render_template
 
 import logging
 log = logging.getLogger("fsa")
@@ -41,12 +42,17 @@ __version__ = pkg.require("FlaskSimpleAuth")[0].version
 
 
 @dataclass
-class FSAException(BaseException):
-    """Exception class to carry fields for an error Response."""
+class ErrorResponse(BaseException):
+    """Internal exception class to carry fields for an error Response."""
     message: str
     status: int
 
 
+class ConfigError(BaseException):
+    """FSA Configuration User Error"""
+    pass
+
+
 #
 # TYPE CASTS
 #
@@ -66,6 +72,16 @@ class JsonData:
     pass
 
 
+class Session:
+    """Session parameter type."""
+    pass
+
+
+class Globals:
+    """Global parameter type."""
+    pass
+
+
 #
 # SPECIAL PREDEFINED GROUP NAMES
 #
@@ -79,44 +95,83 @@ _PREDEFS = (ANY, ALL, NONE)
 
 def _typeof(p: inspect.Parameter):
     """Guess parameter type, possibly with some type inference."""
-    return dict if p.kind is inspect.Parameter.VAR_KEYWORD else \
+    return \
+        dict if p.kind is inspect.Parameter.VAR_KEYWORD else \
         list if p.kind is inspect.Parameter.VAR_POSITIONAL else \
         p.annotation if p.annotation is not inspect._empty else \
         type(p.default) if p.default and p.default is not inspect._empty else \
-        str  # type: ignore
+        str
 
 
+# FIXME probably such a class already exists somewhere and should be reused?
 class _Pool:
-    """Thread-safe pool of something, created on demand."""
+    """Thread-safe pool of something, created on demand.
 
-    def __init__(self, fun: Callable[[int], Any], max_size: int = 0):
-        self._lock = threading.Lock()
+    - fun: function to create objects on demand, called with the creation number.
+    - max_size: maximum size of the pool, 0 for unlimited.
+    - max_use: how many times to use a something, 0 for unlimited.
+    """
+
+    def __init__(self, fun: Callable[[int], Any], max_size: int = 0, max_use: int = 0):
+        self._lock = threading.RLock()
         self._fun = fun
         self._nobjs = 0
+        self._ncreated = 0
         self._max_size = max_size
-        self._available: Set[Any] = set()
+        self._max_use = max_use
+        # pool's content: available vs in use objects
+        self._avail: Set[Any] = set()
         self._using: Set[Any] = set()
+        # keep track of usage
+        self._uses: Dict[Any, int] = dict()
+
+    def __delete__(self):
+        """This should be done automatically, but eventually."""
+        with self._lock:
+            self._using.clear()
+            self._uses.clear()
+            for obj in list(self._avail):
+                if hasattr(obj, "close"):  # pragma: no cover
+                    try:
+                        obj.close()
+                    except Exception:
+                        pass
+                del obj
+            self._avail.clear()
 
     def get(self):
         """Get a object from the pool, possibly creating one if needed."""
         with self._lock:
             try:
-                obj = self._available.pop()
+                obj = self._avail.pop()
                 self._using.add(obj)
-            except KeyError:
+                self._uses[obj] += 1
+            except KeyError:  # nothing available
                 if self._max_size and self._nobjs >= self._max_size:
                     raise Exception(f"object pool max size reached ({self._max_size})")
                 log.debug(f"creating new obj with {self._fun}")
-                obj = self._fun(self._nobjs)
+                obj = self._fun(self._ncreated)
+                self._ncreated += 1
                 self._nobjs += 1
                 self._using.add(obj)
+                self._uses[obj] = 1
             return obj
 
     def ret(self, obj):
         """Return object to pool."""
         with self._lock:
             self._using.remove(obj)
-            self._available.add(obj)
+            if self._max_use and self._uses[obj] >= self._max_use:
+                if hasattr(obj, "close"):
+                    try:
+                        obj.close()
+                    except Exception as e:
+                        log.warning(f"exception occured on close(): {e}")
+                del self._uses[obj]
+                del obj
+                self._nobjs -= 1
+            else:
+                self._avail.add(obj)
 
 
 class Reference:
@@ -138,39 +193,39 @@ class Reference:
     """
 
     class Local(object):
-        """
-        Dumb storage class for shared mode.
-        """
+        """Dumb storage class for shared mode."""
         pass
 
     class Mode(Enum):
-        """
-        Define granularity of object sharing
+        """Granularity of object sharing.
 
         - SHARED: only one object, which should be thread safe.
         - THREAD: per-thread object, generated by a function.
-        - VERSATILE: sub-thread-level object (eg greenlet), generated by a
-          function.
+        - VERSATILE: sub-thread-level object (eg greenlet), generated by a function.
         """
         AUTO = 0
         SHARED = 1
         THREAD = 2
         VERSATILE = 3
 
-    def __init__(self, obj: Any = None, set_name: str = "set", fun: Optional[Callable] = None, max_size: Optional[int] = None, mode: Mode = Mode.AUTO):
+    def __init__(self, obj: Any = None, set_name: str = "set", fun: Optional[Callable] = None,
+                 max_size: int = 0, max_use: int = 0, mode: Mode = Mode.AUTO):
         """Constructor parameters:
 
         - set_name: provide another prefix for the "set" functions.
         - obj: object to be wrapped, can also be provided later.
         - fun: function to generated a per-thread/or-whatever wrapped object.
         - max_size: pool maximum size, 0 for unlimited, None for no pooling.
+        - max_use: when pooling, how many times to reuse an object.
         - mode: level of sharing, default is to chose between SHARED and THREAD.
         """
         # mode encodes the expected object unicity or multiplicity
-        self._mode = Reference.Mode.SHARED if mode == Reference.Mode.AUTO and obj else \
-                     Reference.Mode.THREAD if mode == Reference.Mode.AUTO and fun else \
-                     mode
+        self._mode = \
+            Reference.Mode.SHARED if mode == Reference.Mode.AUTO and obj else \
+            Reference.Mode.THREAD if mode == Reference.Mode.AUTO and fun else \
+            mode
         self._pool_max_size = max_size
+        self._pool_max_use = max_use
         self._set(obj=obj, fun=fun, mandatory=False)
         if set_name and set_name != "_set":
             setattr(self, set_name, self._set)
@@ -179,7 +234,7 @@ class Reference:
 
     def _set_obj(self, obj):
         """Set current wrapped object."""
-        log.debug(f"setting reference to {obj} ({type(obj)})")
+        log.debug(f"Setting Reference to {obj} ({type(obj)})")
         self._mode = Reference.Mode.SHARED
         self._fun = None
         self._pool = None
@@ -194,7 +249,7 @@ class Reference:
             self._mode = Reference.Mode.THREAD
         assert self._mode in (Reference.Mode.THREAD, Reference.Mode.VERSATILE)
         self._fun = fun
-        self._pool = _Pool(fun, self._pool_max_size) \
+        self._pool = _Pool(fun, self._pool_max_size, self._pool_max_use) \
             if self._pool_max_size is not None else None
         self._nobjs = 0
         if self._mode == Reference.Mode.THREAD:
@@ -207,23 +262,26 @@ class Reference:
     def _set(self, obj: Any = None, fun: Optional[Callable[[int], Any]] = None, mandatory=True):
         """Set current wrapped object or generation function."""
         if obj and fun:
-            raise Exception("reference cannot set both obj and fun")
+            raise Exception("Reference cannot set both obj and fun")
         elif obj:
             return self._set_obj(obj)
         elif fun:
             return self._set_fun(fun)
         elif mandatory:
-            raise Exception("reference must set either obj or fun")
+            raise Exception("Reference must set either obj or fun")
 
     def _get_obj(self):
-        """Get current wrapped object."""
+        """Get current wrapped object, possibly creating it."""
+        # handle creation
         if self._fun and not hasattr(self._local, "obj"):
             if self._pool_max_size is not None:
-                self._local.obj = self._pool.get()
-                self._nobjs = self._pool._nobjs  # FIXME
-            else:
+                # sync on pool to extract a consistent nobjs
+                with self._pool._lock:
+                    self._local.obj = self._pool.get()
+                    self._nobjs = self._pool._nobjs
+            else:  # no pool
                 self._local.obj = self._fun(self._nobjs)
-                self._nobjs += 1  # FIXME?
+                self._nobjs += 1
         return self._local.obj
 
     # FIXME how to do that automatically when the thread/whatever ends?
@@ -286,19 +344,21 @@ class Flask(flask.Flask):
         self.get_user = self._fsa.get_user
         self.current_user = self._fsa.current_user
         # overwrite decorators ("route" done through add_url_rule above)
-        setattr(self, "get", self._fsa.get)  # avoid mypy warnings
+        setattr(self, "get", self._fsa.get)  # FIXME avoid mypy warnings
         setattr(self, "put", self._fsa.put)
         setattr(self, "post", self._fsa.post)
         setattr(self, "patch", self._fsa.patch)
         setattr(self, "delete", self._fsa.delete)
 
 
-# all possible directives
+# all possible flask-simple-auth directives
 _DIRECTIVES = {
     # debug
     "FSA_DEBUG", "FSA_LOGGING_LEVEL",
     # general settings
     "FSA_SECURE", "FSA_SERVER_ERROR", "FSA_NOT_FOUND_ERROR",
+    # parameter handing
+    "FSA_REJECT_UNEXPECTED_PARAM",
     # register hooks
     "FSA_GET_USER_PASS", "FSA_USER_IN_GROUP", "FSA_CAST", "FSA_OBJECT_PERMS",
     # authentication
@@ -319,6 +379,7 @@ _DIRECTIVES = {
 _DEFAULT_CACHE = "ttl"
 _DEFAULT_CACHE_SIZE = 262144  # a few MB
 _DEFAULT_CACHE_TTL = 600  # seconds, 10 minutes
+_DEFAULT_REJECT_UNEXPECTED_PARAM = True
 _DEFAULT_SERVER_ERROR = 500
 _DEFAULT_NOT_FOUND_ERROR = 404
 _DEFAULT_PASSWORD_SCHEME = "bcrypt"
@@ -342,8 +403,7 @@ class FlaskSimpleAuth:
         self._casts: Dict[type, Callable[[str], object]] = {
             bool: lambda s: None if s is None else s.lower() not in ("", "0", "false", "f"),
             int: lambda s: int(s, base=0) if s else None,
-            # NOTE mypy complains wrongly about non-existing _empty.
-            inspect._empty: str,  # type: ignore
+            inspect._empty: str,
             path: str,
             string: str,
             dt.date: dt.date.fromisoformat,
@@ -371,15 +431,15 @@ class FlaskSimpleAuth:
         return Response(msg, code, content_type="text/plain")
 
     def _Err(self, msg: str, code: int):
-        """Build and trace an FSAException."""
+        """Build and trace an ErrorResponse."""
         if self._debug:
-            log.debug(f"fsa exception: {code} {msg}")
-        return FSAException(msg, code)
+            log.debug(f"error: {code} {msg}")
+        return ErrorResponse(msg, code)
 
     def _Bad(self, msg: str):
         """Build and trace an exception on a bad configuration."""
         log.critical(msg)
-        return Exception(msg)
+        return ConfigError(msg)
 
     def _auth_has(self, *auth):
         """Tell whether current authentication includes any of these schemes."""
@@ -423,8 +483,8 @@ class FlaskSimpleAuth:
             location = self._401_redirect
             # allow to come back later in some cases
             if self._url_name and request.method == "GET":
-                sep = "&" if "?" in self._url_name else "?"
                 import urllib
+                sep = "&" if "?" in self._url_name else "?"
                 location += sep + urllib.parse.urlencode({self._url_name: request.url})
             return redirect(location, 307)
         return res
@@ -443,6 +503,7 @@ class FlaskSimpleAuth:
                 else:
                     set_cookie = True
                 if set_cookie:
+                    # path? other parameters?
                     res.set_cookie(self._name, self.create_token(self._local.user),
                                    max_age=int(60 * self._delay))
         return res
@@ -491,7 +552,7 @@ class FlaskSimpleAuth:
         return uig
 
     def cast(self, t, cast: Optional[Callable] = None):
-        """Add a cast function to a type."""
+        """Add a cast function associated to a type."""
         if t in self._casts:
             log.warning(f"overriding type casting function for {t}")
         if cast:  # direct
@@ -503,7 +564,7 @@ class FlaskSimpleAuth:
             return annotate
 
     def object_perms(self, domain, checker: Optional[Callable] = None):
-        """Add an object permission helper for a domain."""
+        """Add an object permission helper for a given domain."""
         if domain in self._object_perms:
             log.warning(f"overriding object permission checker for domain {domain}")
         if checker:  # direct
@@ -550,12 +611,14 @@ class FlaskSimpleAuth:
         # check directives for typos
         for name in conf:
             if name.startswith("FSA_") and name not in _DIRECTIVES:
-                log.warning(f"unexpected directive: {name}")
+                log.warning(f"unexpected directive, ignored: {name}")
         # whether to only allow secure requests
         self._secure = conf.get("FSA_SECURE", True)
         # status code for some errors errors
         self._server_error = conf.get("FSA_SERVER_ERROR", _DEFAULT_SERVER_ERROR)
         self._not_found_error = conf.get("FSA_NOT_FOUND_ERROR", _DEFAULT_NOT_FOUND_ERROR)
+        # whether to error on unexpected parameters
+        self._reject_param = conf.get("FSA_REJECT_UNEXPECTED_PARAM", _DEFAULT_REJECT_UNEXPECTED_PARAM)
         #
         # overall auth setup
         #
@@ -570,7 +633,8 @@ class FlaskSimpleAuth:
         else:
             self._auth = auth
         for a in self._auth:
-            assert a in self._FSA_AUTH
+            if a not in self._FSA_AUTH:
+                raise self._Bad(f"unexpected auth: {a}")
         self._local.auth = self._auth
         #
         # web apps…
@@ -683,7 +747,7 @@ class FlaskSimpleAuth:
             # list of 94 chars, about 6.5 bits per char, 40 chars => 260 bits
             chars = string.ascii_letters + string.digits + string.punctuation
             self._secret = "".join(random.SystemRandom().choices(chars, k=40))
-        if not self._token:
+        if not self._token:  # pragma: no cover
             pass
         elif self._token == "fsa":
             self._sign = self._secret
@@ -783,8 +847,9 @@ class FlaskSimpleAuth:
         #
         self.blueprints = self._app.blueprints
         self.debug = False
-        if hasattr(self._app, '_is_setup_finished'):  # Flask 2.0
-            self._is_setup_finished = self._app._is_setup_finished
+        if hasattr(self._app, '_check_setup_finished'):
+            # Flask 2.2
+            self._check_setup_finished = self._app._check_setup_finished
             self.before_request_funcs = self._app.before_request_funcs
             self.after_request_funcs = self._app.after_request_funcs
             self.teardown_request_funcs = self._app.teardown_request_funcs
@@ -843,7 +908,7 @@ class FlaskSimpleAuth:
         params = self._params()
         user = params.get(self._login, None)
         if not user:
-            raise self._Err("missing login parameter", 401)
+            raise self._Err(f"missing login parameter: {self._login}", 401)
         return user
 
     #
@@ -878,8 +943,8 @@ class FlaskSimpleAuth:
         Raise an exception if not ok, otherwise simply proceeds."""
         try:
             ref = self._get_user_pass(user)
-        except FSAException as fe:
-            raise fe
+        except ErrorResponse as e:
+            raise e
         except Exception as e:
             log.error(f"get_user_pass failed: {e}")
             raise self._Err("internal error in get_user_pass", self._server_error)
@@ -908,9 +973,9 @@ class FlaskSimpleAuth:
             user = self._http_auth.authenticate(auth, password)
             if user is not None and user is not False:
                 return auth.username if user is True else user
-        except FSAException as error:
-            log.debug(f"AUTH (http-*): bad authentication {error}")
-            raise error
+        except ErrorResponse as e:
+            log.debug(f"AUTH (http-*): bad authentication {e}")
+            raise e
         log.debug("AUTH (http-*): bad authentication")
         raise self._Err("failed HTTP authentication", 401)
 
@@ -962,7 +1027,7 @@ class FlaskSimpleAuth:
         """Get user from basic or param authentication."""
         try:
             return self._get_basic_auth()
-        except FSAException:  # failed, let's try param
+        except ErrorResponse:  # failed, let's try param
             return self._get_param_auth()
 
     #
@@ -1151,7 +1216,7 @@ class FlaskSimpleAuth:
                 self._local.user = self._FSA_AUTH[a](self)
                 if self._local.user:
                     break
-            except FSAException as e:
+            except ErrorResponse as e:
                 lae = e
             except Exception as e:  # pragma: no cover
                 log.error(f"internal error in {a} authentication: {e}")
@@ -1161,7 +1226,7 @@ class FlaskSimpleAuth:
 
         # rethrow last auth exception on failure
         if required and not self._local.user:
-            raise lae or FSAException("missing authentication", 401)
+            raise lae or self._Err("missing authentication", 401)
 
         return self._local.user
 
@@ -1208,7 +1273,7 @@ class FlaskSimpleAuth:
         """Call a route function ensuring a response whatever."""
         try:  # the actual call
             return fun(*args, **kwargs)
-        except FSAException as e:  # something went wrong
+        except ErrorResponse as e:  # something went wrong
             return self._Res(e.message, e.status)
         except Exception as e:  # something went really wrong
             log.error(f"internal error on {request.method} {request.path}: {e}")
@@ -1238,7 +1303,7 @@ class FlaskSimpleAuth:
                         self._local.auth = auth
                     try:
                         self._local.user = self.get_user()
-                    except FSAException as e:
+                    except ErrorResponse as e:
                         return self._Res(e.message, e.status)
 
                 if not self._local.user:  # pragma no cover
@@ -1254,10 +1319,11 @@ class FlaskSimpleAuth:
         """Decorator to authorize user groups."""
 
         for group in _PREDEFS:
-            assert group not in groups, f"unexpected predefined {group}"
+            if group in groups:
+                raise self._Bad(f"unexpected predefined {group}")
 
-        assert self._user_in_group, \
-            "user_in_group callback needed for group authorization on {path}"
+        if not self._user_in_group:  # pragma: no cover
+            raise self._Bad(f"user_in_group callback needed for group authorization on {path}")
 
         def decorate(fun: Callable):
 
@@ -1271,8 +1337,8 @@ class FlaskSimpleAuth:
                 for group in groups:
                     try:
                         ok = self._user_in_group(self._local.user, group)
-                    except FSAException as fe:
-                        return self._Res(fe.message, fe.status)
+                    except ErrorResponse as e:
+                        return self._Res(e.message, e.status)
                     except Exception as e:
                         log.error(f"user_in_group failed: {e}")
                         return self._Res("internal error in user_in_group", self._server_error)
@@ -1303,6 +1369,7 @@ class FlaskSimpleAuth:
             # parameters types/casts and defaults taken from signature
             sig, keywords = inspect.signature(fun), False
 
+            # build helpers
             for n, p in sig.parameters.items():
                 sn = n[1:] if n[0] == '_' and len(n) > 1 else n
                 names[n], names[sn] = sn, n
@@ -1312,7 +1379,7 @@ class FlaskSimpleAuth:
                     t = _typeof(p)
                     types[n] = t
                     typings[n] = self._casts.get(t, t)
-                if p.default != inspect._empty:  # type: ignore
+                if p.default != inspect._empty:
                     defaults[n] = p.default
                 if p.kind == p.VAR_KEYWORD:
                     keywords = True
@@ -1342,6 +1409,12 @@ class FlaskSimpleAuth:
                         else:
                             if p in defaults:
                                 kwargs[p] = defaults[p]
+                            elif typing == Request:
+                                kwargs[p] = request
+                            elif typing == Session:
+                                kwargs[p] = session
+                            elif typing == Globals:
+                                kwargs[p] = g
                             else:
                                 return self._Res(f"missing parameter \"{pn}\"", 400)
                     else:
@@ -1357,10 +1430,14 @@ class FlaskSimpleAuth:
                     for p in params:
                         if p not in kwargs:
                             kwargs[p] = params[p]
-                elif self._debug:  # warn about unused parameters
+                elif self._debug or self._reject_param:
+                    # detect unused parameters and warn or reject them
                     for p in params:
                         if p not in names and p not in self._names:
-                            log.debug(f"unexpected parameter {p} on {path}")
+                            if self._debug:
+                                log.debug(f"unexpected parameter {p} on {path}")
+                            if self._reject_param:
+                                return self._Res(f"unexpected parameter {p} on {path}", 400)
 
                 return self._safe_call(path, "parameters", fun, *args, **kwargs)
 
@@ -1407,7 +1484,7 @@ class FlaskSimpleAuth:
 
                     try:
                         ok = self._check_object_perms(self._local.user, domain, val, mode)
-                    except FSAException as e:
+                    except ErrorResponse as e:
                         return self._Res(e.message, e.status)
                     except Exception as e:
                         log.error(f"internal error on {request.method} {request.path} permission {perm} check: {e}")
@@ -1464,7 +1541,7 @@ class FlaskSimpleAuth:
         if len(authorize) == 0:
             authorize = [NONE]
 
-        # separate groups and perms
+        # separate predefs, groups and perms
         predefs = list(filter(lambda a: a in _PREDEFS, authorize))
         groups = list(filter(lambda a: type(a) in (int, str) and a not in _PREDEFS, authorize))
         perms = list(filter(lambda a: isinstance(a, tuple), authorize))
@@ -1486,13 +1563,35 @@ class FlaskSimpleAuth:
 
         from uuid import UUID
         # add the expected type to path sections, if available
-        # flask converter types: string (default), int, float, path, uuid
+        # flask converters: string (default), int, float, path, uuid
+        # NOTE it can be extended (`url_map`), but we are managing through annotations
         sig = inspect.signature(view_func)
 
         splits = rule.split("<")
         for i, s in enumerate(splits):
             if i > 0:
                 spec, remainder = s.split(">", 1)
+                # some sanity checks on path parameters
+                conv, name = spec.split(":") if ":" in spec else (None, spec)
+                # check for predefined converters, but "any"
+                if conv not in (None, "string", "int", "float", "path", "uuid"):
+                    raise self._Bad(f"unexpected converter: {conv}")
+                if name in sig.parameters:
+                    namesig = sig.parameters[name]
+                    if namesig.default != inspect._empty:
+                        raise self._Bad(f"path parameter cannot have a default: {name}")
+                    if conv and namesig.annotation != inspect.Parameter.empty:
+                        atype = namesig.annotation
+                        if conv == "string" and atype not in (string, str) or \
+                           conv == "uuid" and atype != UUID or \
+                           conv == "path" and atype not in (path, str) or \
+                           conv == "int" and atype != int or \
+                           conv == "float" and atype != float:
+                            raise self._Bad(f"inconsistent type for {conv} converter on {name}: {atype}")
+                    # else no annotation consistency to check
+                else:
+                    raise self._Bad(f"path parameter missing from function signature: {name}")
+                # add explicit "Flask" path parameter type
                 if ":" not in spec and spec in sig.parameters:
                     t = _typeof(sig.parameters[spec])
                     # Flask supports 5 types, with string the default?
@@ -1500,6 +1599,7 @@ class FlaskSimpleAuth:
                         splits[i] = f"{t.__name__.lower()}:{spec}>{remainder}"
                     else:
                         splits[i] = f"string:{spec}>{remainder}"
+                # else spec includes a type that we keep…
         newpath = "<".join(splits)
 
         # special shortcut for NONE
@@ -1556,7 +1656,7 @@ class FlaskSimpleAuth:
         return decorate
 
     # support Flask 2.0 per-method decorator shortcuts
-    # note that app.get("/", methods=["POST"], ...) would do a POST.
+    # note that app.get("/", methods=["POST"], …) would do a POST.
     def get(self, rule, **options):
         """Shortcut for `route` with `GET` method."""
         return self.route(rule, methods=["GET"], **options)
