@@ -12,7 +12,8 @@ log = logging.getLogger("app")
 # APP
 #
 import FlaskSimpleAuth as fsa
-from FlaskSimpleAuth import Flask, jsonify, ANY, ALL, NONE, path, string
+from FlaskSimpleAuth import Flask, jsonify, ANY, ALL, NONE, path, string, \
+    Request, Session, Globals
 app = Flask("Test", debug=True)
 
 #
@@ -332,7 +333,7 @@ def get_my_login(login: str):
 # raise instead of return
 @app.get("/oops", authorize=ANY)
 def get_oops():
-    raise fsa.FSAException("Ooops!", 518)
+    raise fsa.ErrorResponse("Ooops!", 518)
 
 # magic Json handling
 import json as js
@@ -340,3 +341,8 @@ from FlaskSimpleAuth import JsonData
 @app.get("/json", authorize=ANY)
 def get_json(j: JsonData):
     return f"{type(j).__name__}: {js.dumps(j)}", 200
+
+# magic parameters
+@app.get("/request", authorize=ANY)
+def get_request(req: Request, ses: Session, glo: Globals):
+    return req.base_url, 200
