@@ -8,7 +8,7 @@ import App
 from App import app
 
 import FlaskSimpleAuth as fsa
-from FlaskSimpleAuth import Response
+from FlaskSimpleAuth import Response, ConfigError
 import json
 
 import AppExt
@@ -230,7 +230,7 @@ def test_fsa_token():
     try:
         user = app._fsa._get_any_token_auth("R:U:demain:signature")
         assert False, "expecting a bad timestamp format"
-    except fsa.FSAException as e:
+    except fsa.ErrorResponse as e:
         assert "unexpected timestamp format" in e.message
     # force expiration
     grace = app._fsa._grace
@@ -238,14 +238,14 @@ def test_fsa_token():
     try:
         user = app._fsa._get_any_token_auth(calvin_token)
         assert False, "token must have expired"
-    except fsa.FSAException as e:
+    except fsa.ErrorResponse as e:
         assert "expired auth token" in e.message
     # again after clear cache, so the expiration is detected at fsa level
     app.clear_caches()
     try:
         user = app._fsa._get_any_token_auth(calvin_token)
         assert False, "token must have expired"
-    except fsa.FSAException as e:
+    except fsa.ErrorResponse as e:
         assert "expired fsa auth token" in e.message
     # cleanup
     app._fsa._grace = grace
@@ -255,7 +255,7 @@ def test_fsa_token():
     try:
         user = app._fsa._get_any_token_auth(hobbes_token)
         assert False, "token should be invalid"
-    except fsa.FSAException as e:
+    except fsa.ErrorResponse as e:
         assert e.status == 401
     app._fsa._grace = grace
 
@@ -297,7 +297,7 @@ def test_jwt_token():
     try:
         user = app._fsa._get_any_token_auth(susie_token)
         assert False, "expired token should fail"
-    except fsa.FSAException as e:
+    except fsa.ErrorResponse as e:
         assert "expired jwt auth token" in e.message
     finally:
         app._fsa._delay, app._fsa._grace = delay, grace
@@ -314,7 +314,7 @@ def test_jwt_token():
         bad_token = f"{pieces[0]}.{pieces[2]}.{pieces[1]}"
         user = app._fsa._get_any_token_auth(bad_token)
         assert False, "bad token should fail"
-    except fsa.FSAException as e:
+    except fsa.ErrorResponse as e:
         assert "invalid jwt token" in e.message
     # cleanup
     app._fsa._token, app._fsa._algo = tsave, hsave
@@ -327,7 +327,7 @@ def test_invalid_token():
     try:
         user = app._fsa._get_any_token_auth(susie_token)
         assert False, "token should be invalid"
-    except fsa.FSAException as e:
+    except fsa.ErrorResponse as e:
         assert e.status == 401
     # wrong token
     realm, app._fsa._realm = app._fsa._realm, "elsewhere"
@@ -336,7 +336,7 @@ def test_invalid_token():
     try:
         user = app._fsa._get_any_token_auth(moe_token)
         assert False, "token should be invalid"
-    except fsa.FSAException as e:
+    except fsa.ErrorResponse as e:
         assert e.status == 401
 
 def test_password_check(client):
@@ -378,7 +378,7 @@ def test_authorize():
         def foo():
             return "foo", 200
         assert False, "cannot mix ALL & ANY in authorize"
-    except Exception as e:
+    except ConfigError as e:
         assert True, "mix is forbidden"
 
 def test_self_care(client):
@@ -399,10 +399,17 @@ def test_self_care(client):
 def test_typed_params(client):
     res = check(200, client.get("/add/2", data={"a":"2.0", "b":"4.0"}))
     assert float(res.data) == 12.0
+    # unused params are ignored
+    app._fsa._reject_param = False
     res = check(200, client.get("/mul/2", data={"j":"3", "k":"4", "unused":"x"}))
     assert int(res.data) == 24
     res = check(200, client.get("/mul/2", json={"j":5, "k":"4", "unused":"y"}))
     assert int(res.data) == 40
+    # unused params are rejected
+    app._fsa._reject_param = True
+    res = check(400, client.get("/mul/2", data={"j":"3", "k":"4", "unused":"x"}))
+    res = check(400, client.get("/mul/2", json={"j":5, "k":"4", "unused":"y"}))
+    # type errors
     check(400, client.get("/mul/1", data={"j":"3"}))
     check(400, client.get("/mul/1", data={"k":"4"}))
     check(400, client.get("/mul/2", data={"j":"three", "k":"four"}))
@@ -419,6 +426,8 @@ def test_typed_params(client):
     check(400, client.get("/sub", data={"j":"42"}))
     res = check(200, client.get("/sub", data={"i":"42"}))
     assert int(res.data) == 42
+    res = check(200, client.get("/request"))
+    assert res.data.endswith(b"/request")
 
 def test_types(client):
     res = check(200, client.get("/type", data={"f": "1.0"}))
@@ -666,7 +675,7 @@ def test_reference():
     # thread local stuff
     def gen_data(i):
         return f"data: {i}"
-    r = fsa.Reference(fun=gen_data)
+    r = fsa.Reference(fun=gen_data, max_size=None)
     assert r._nobjs == 0
     assert isinstance(r.__hash__(), int)
     assert r._nobjs == 1
@@ -692,13 +701,13 @@ def test_reference():
         r = fsa.Reference(obj="hello", fun=gen_data)
         assert False, "should have raised an exception"
     except Exception as e:
-        assert "reference cannot set both obj and fun" in str(e)
+        assert "Reference cannot set both obj and fun" in str(e)
     try:
         r = fsa.Reference(set_name="add")
         r.add()
         assert False, "missing parameter in previous call"
     except Exception as e:
-        assert "reference must set either obj or fun" in str(e)
+        assert "Reference must set either obj or fun" in str(e)
     # auto thread
     r = fsa.Reference()
     r._set_fun(lambda i: i)
@@ -712,19 +721,19 @@ def test_reference():
 def test_ref_pool():
     ref = fsa.Reference(fun=lambda i: i, max_size=2)
     i = ref._get_obj()
-    assert len(ref._pool._available) == 0
+    assert len(ref._pool._avail) == 0
     assert len(ref._pool._using) == 1
     assert ref._pool._nobjs == 1
     ref._ret_obj()
     i = ref._get_obj()
-    assert len(ref._pool._available) == 0
+    assert len(ref._pool._avail) == 0
     assert len(ref._pool._using) == 1
     assert ref._pool._nobjs == 1
     # this should return the same object as we are in the same thread
     j = ref._get_obj()
     assert i == j
     ref._ret_obj()
-    assert len(ref._pool._available) == 1
+    assert len(ref._pool._avail) == 1
     assert len(ref._pool._using) == 0
     assert ref._pool._nobjs == 1
     # test with 2 ordered threads to grasp to objects from the pool
@@ -754,6 +763,38 @@ def test_ref_pool():
         assert "pool max size reached" in str(e)
 
 
+def test_pool():
+    # test max_use
+    p1 = fsa._Pool(fun = lambda i: i, max_size = 1, max_use = 2)
+    i = p1.get()
+    assert i == 0
+    p1.ret(i)
+    i = p1.get()
+    assert i == 0
+    p1.ret(i)
+    i = p1.get()
+    assert i == 1
+    p1.ret(i)
+    p1.__delete__()
+    # test with close and None
+    class T:
+        def __init__(self, count):
+            self._count = count
+        def close(self):
+            self._count = None
+            raise Exception("Oops!")
+        def __str__(self):
+            return f"T({self._count})"
+    p2 = fsa._Pool(fun = T, max_size = None, max_use = 1)
+    t = p2.get()
+    assert str(t) == "T(0)"
+    p2.ret(t)
+    t = p2.get()
+    assert str(t) == "T(1)"
+    p2.ret(t)
+    p2.__delete__()
+
+
 def test_www_authenticate(client):
     push_auth(app._fsa, "param")
     res = check(401, client.get("/admin"))
@@ -897,57 +938,58 @@ def test_bad_app():
     try:
         app = create_app(FSA_AUTH="bad")
         assert False, "bad app creation must fail"
-    except Exception:
-        assert True, "ok, bad app creation has failed"
+    except ConfigError as e:
+        assert "bad" in str(e), "ok, bad app creation has failed"
     try:
         app = create_app(FSA_AUTH=["fake", "basic", "bad"])
         assert False, "bad app creation must fail"
-    except Exception:
-        assert True, "ok, bad app creation has failed"
+    except ConfigError as e:
+        assert "bad" in str(e), "ok, bad app creation has failed"
     try:
         app = create_app(auth="bad")
         assert False, "bad app creation must fail"
-    except Exception:
-        assert True, "ok, bad app creation has failed"
+    except ConfigError as e:
+        assert "bad" in str(e), "ok, bad app creation has failed"
     try:
         app = create_app(auth=["basic", "token", "bad"])
         assert False, "bad app creation must fail"
-    except Exception:
-        assert True, "ok, bad app creation has failed"
+    except ConfigError as e:
+        assert "bad" in str(e), "ok, bad app creation has failed"
     # bad token type
     try:
         app = create_app(FSA_AUTH="token", FSA_TOKEN_TYPE="bad")
         assert False, "bad app creation must fail"
-    except Exception:
-        assert True, "ok, bad app creation has failed"
-    try:
-        app = create_app(FSA_AUTH="token", FSA_TOKEN_TYPE=None)
-        assert False, "bad app creation must fail"
-    except Exception:
-        assert True, "ok, bad app creation has failed"
+    except ConfigError as e:
+        assert "bad" in str(e), "ok, bad app creation has failed"
+    # FIXME, None is ok?
+    # try:
+    #     app = create_app(FSA_AUTH="token", FSA_TOKEN_TYPE=None)
+    #    # assert False, "bad app creation must fail"
+    # except ConfigError as e:
+    #     assert True, "ok, bad app creation has failed"
     # bad token carrier
     try:
         app = create_app(FSA_AUTH="token", FSA_TOKEN_CARRIER="bad")
         assert False, "bad app creation must fail"
-    except Exception:
-        assert True, "ok, bad app creation has failed"
+    except ConfigError as e:
+        assert "bad" in str(e), "ok, bad app creation has failed"
     try:
         app = create_app(FSA_AUTH="token", FSA_TOKEN_CARRIER=None)
         assert False, "bad app creation must fail"
-    except Exception:
-        assert True, "ok, bad app creation has failed"
+    except ConfigError as e:
+        assert "requires a carrier" in str(e), "ok, bad app creation has failed"
     # bad token name
     try:
         app = create_app(FSA_AUTH="token", FSA_TOKEN_CARRIER="bearer", FSA_TOKEN_NAME=None)
         assert False, "bad app creation must fail"
-    except Exception:
-        assert True, "ok, bad app creation has failed"
+    except ConfigError as e:
+        assert "requires a name" in str(e), "ok, bad app creation has failed"
     # bad jwt talgorithm
     try:
         app = create_app(FSA_AUTH="token", FSA_TOKEN_TYPE="jwt", FSA_TOKEN_ALGO="bad")
         assert False, "bad app creation must fail"
-    except Exception:
-        assert True, "ok, bad app creation has failed"
+    except ConfigError as e:
+        assert "bad" in str(e), "ok, bad app creation has failed"
 
 class PK():
     def __init__(self, kind):
@@ -1029,6 +1071,30 @@ def test_bad_4(bad4):
     assert b"internal error caught at no authorization on /any" == res.data
     check(404, bad4.get("/no-such-route"))
 
+
+def test_bads():
+    import AppBad as ab
+    try:
+        app = ab.create_badapp_5()
+        assert False, "mandatory parameter with default should fail"
+    except ConfigError as e:
+        assert "default" in str(e)
+    try:
+        app = ab.create_badapp_6()
+        assert False, "missing path parameter should fail"
+    except ConfigError as e:
+        assert "missing" in str(e)
+    try:
+        app = ab.create_badapp_7()
+        assert False, "inconsistent path parameter types should fail"
+    except ConfigError as e:
+        assert "bad" in str(e)
+    try:
+        app = ab.create_badapp_8()
+        assert False, "unknown path parameter converter should fail"
+    except ConfigError as e:
+        assert "unknown" in str(e)
+
 # per-object perms
 def test_object_perms(client):
     check(401, client.get("/my/calvin", data={"LOGIN": None}))
@@ -1052,7 +1118,7 @@ def test_object_perms_errors():
     def is_okay(u: str, v: str, m: str):
         log.debug(f"is_okay({u}, {v}, {m})")
         if v == "fsa":
-            raise fsa.FSAException("oops-1", 518)
+            raise fsa.ErrorResponse("oops-1", 518)
         elif v == "ex":
             raise Exception("oops-2")
         elif v == "float":
@@ -1067,42 +1133,42 @@ def test_object_perms_errors():
         def get_bad_perm_1(uid: int):
             return "should not get there", 200
         assert False, "should detect too short tuple"
-    except Exception as e:
+    except ConfigError as e:
         assert "3 data" in str(e)
     try:
         @app.get("/bad-perm-2/<uid>", authorize=("unknown",))
         def get_bad_perm_2_uid(uid: int):
             return "should not get there", 200
         assert False, "should detect unregistered permission domain"
-    except Exception as e:
+    except ConfigError as e:
         assert "missing object permission" in str(e)
     try:
         @app.get("/bad-perm-3", authorize=("known", 3))
         def get_bad_perm_3(uid: int):
             return "should not get there", 200
         assert False, "should detect bad variable name"
-    except Exception as e:
+    except ConfigError as e:
         assert "unexpected identifier name type" in str(e)
     try:
         @app.get("/bad-perm-4/<uid>", authorize=("known", "uid", 3.14159))
         def get_bad_perm_4_uid(uid: int):
             return "should not get there", 200
         assert False, "should detect bad mode type"
-    except Exception as e:
+    except ConfigError as e:
         assert "unexpected mode type" in str(e)
     try:
         @app.get("/bad-perm-5", authorize=("known", "uid"))
         def get_bad_perm_3(oid: int):
             return "should not get there", 200
         assert False, "should detect missing variable"
-    except Exception as e:
+    except ConfigError as e:
         assert "missing function parameter uid" in str(e)
     try:
         @app.get("/bad-perm-6", authorize=("known", "uid"))
         def get_bad_perm_3():
             return "should not get there", 200
         assert False, "should detect missing variable"
-    except Exception as e:
+    except ConfigError as e:
         assert "permissions require some parameters" in str(e)
     # run time errors
     @app.get("/oops/<err>", authorize=("known", "err"))
@@ -1124,35 +1190,35 @@ def test_authorize_errors():
         def get_bad_authorize():
             return "should not get there", 200
         assert False, "should detect bad authorize type"
-    except Exception as e:
+    except ConfigError as e:
         assert "unexpected authorization" in str(e)
     try:
         @app.get("/bad-mix-1", authorize=["ANY", "ALL"])
         def get_bad_mix_1():
             return "should not get there", 200
         assert False, "should detect ANY/ALL mix"
-    except Exception as e:
+    except ConfigError as e:
         assert "ANY/ALL" in str(e)
     try:
         @app.get("/bad-mix-2", authorize=["ANY", "OTHER"])
         def get_bad_mix_2():
             return "should not get there", 200
         assert False, "should detect ANY/other mix"
-    except Exception as e:
+    except ConfigError as e:
         assert "other" in str(e)
     try:
         @app.get("/bad-mix-3", authorize=["ANY", ("foo", "id")])
         def get_bad_mix_2():
             return "should not get there", 200
         assert False, "should detect ANY/other mix"
-    except Exception as e:
+    except ConfigError as e:
         assert "object" in str(e)
 
 def test_group_errors():
     import AppFact as af
     def bad_uig(login, group):
        if group == "ex":
-           raise fsa.FSAException("exception in user_in_group", 518)
+           raise fsa.ErrorResponse("exception in user_in_group", 518)
        elif group == "float":
            return 3.14159
        else:
@@ -1225,12 +1291,12 @@ def test_no_such_cache():
     try:
         af.create_app(FSA_CACHE="no-such-cache")
         assert False, "create app should fail"
-    except Exception as e:
+    except ConfigError as e:
         assert "unexpected FSA_CACHE" in e.args[0]
 
 def test_warnings_and_errors():
     def bad_gup_1(user: str):
-        raise fsa.FSAException("bad_gup_1", 518)
+        raise fsa.ErrorResponse("bad_gup_1", 518)
     def bad_gup_2(user: str):
         return 3.14159
     import AppFact as af
@@ -1252,17 +1318,17 @@ def test_warnings_and_errors():
     try:
         app.hash_password("short1")
         assert False, "should be too short"
-    except fsa.FSAException as e:
+    except fsa.ErrorResponse as e:
         assert e.status == 400 and "too short" in e.message
     try:
         app.hash_password("long-enough-but-missing-a-number")
         assert False, "should not match re"
-    except fsa.FSAException as e:
+    except fsa.ErrorResponse as e:
         assert e.status == 400 and "must match" in e.message
     try:
         app._fsa._check_password("calvin", "hobbes")
         assert False, "should not get through"
-    except fsa.FSAException as e:
+    except fsa.ErrorResponse as e:
         assert e.status == 518 and "bad" in e.message
     # unused length warning
     app = af.create_app(
@@ -1274,7 +1340,7 @@ def test_warnings_and_errors():
     try:
         app._fsa._check_password("calvin", "hobbes")
         assert False, "should not get through"
-    except fsa.FSAException as e:
+    except fsa.ErrorResponse as e:
         assert e.status == 500 and "internal error with get_user_pass" in e.message
     # overwrite warning
     @app.cast("foo")
@@ -1289,12 +1355,12 @@ def test_warnings_and_errors():
     try:
         app = af.create_app(FSA_CAST="not a dict")
         assert False, "should not get through"
-    except Exception as e:
+    except ConfigError as e:
         assert "FSA_CAST must be a dict" in str(e)
     try:
         app = af.create_app(FSA_OBJECT_PERMS="should be a dict")
         assert False, "should not get through"
-    except Exception as e:
+    except ConfigError as e:
         assert "FSA_OBJECT_PERMS must be a dict" in str(e)
 
 def test_jsondata(client):
